@@ -1,4 +1,4 @@
-;; mailcrypt.el v3.5.3, mail encryption with PGP
+;; mailcrypt.el v3.5.4, mail encryption with PGP
 ;; Copyright (C) 1995  Jin Choi <jin@atype.com>
 ;;                     Patrick LoPresti <patl@lcs.mit.edu>
 ;;           (C) 1998  Len Budney <lbudney@pobox.com>
@@ -33,6 +33,7 @@
 
 (require 'easymenu)
 (require 'comint)
+(require 'rfc822)
 
 (eval-and-compile
   (condition-case nil (require 'itimer) (error nil))
@@ -202,7 +203,7 @@
 ;;}}}
 
 ;;{{{ User variables.
-(defconst mc-version "3.5.3")
+(defconst mc-version "3.5.4")
 (defvar mc-temp-directory "/tmp"
   "*Default temp directory to be used by Mailcrypt.")
 (defvar mc-default-scheme 'mc-scheme-pgp "*Default encryption scheme to use.")
@@ -397,7 +398,7 @@ Optional arg NUKE, if non-nil, means eliminate all fields returned."
 
 (defsubst mc-strip-address (addr)
   "Strip everything from ADDR except the basic Email address."
-  (car (cdr (mail-extract-address-components addr))))
+  (car (rfc822-addresses addr)))
 
 (defun mc-strip-addresses (addr-list)
   "Strip everything from the addresses in ADDR-LIST except the basic
@@ -405,7 +406,7 @@ Email address.  ADDR-LIST may be a single string or a list of strings."
   (if (not (listp addr-list)) (setq addr-list (list addr-list)))
   (setq addr-list
 	(mapcar
-	 (function (lambda (s) (mc-split "\\([ \t\n]*,[ \t\n]*\\)" s)))
+	 (function (lambda (s) (rfc822-addresses s)))
 	 addr-list))
   (setq addr-list (apply 'append addr-list))
   (mapconcat 'mc-strip-address addr-list ", "))
@@ -430,9 +431,7 @@ Email address.  ADDR-LIST may be a single string or a list of strings."
 		     (match-beginning 0) (match-end 0))
 		  (setq retval nil)
 		  default))))
-    (if msg (if window-system
-		(progn (message nil) (message-box "%s" msg))
-	      (message "%s" msg)))
+    (if msg (message-or-box "%s" msg))
     retval))
 
 (defun mc-process-region (beg end passwd program args parser &optional buffer)
